import argparse
import asyncio
import logging
import signal
from asyncio import Event
from pathlib import Path
from types import FrameType
from typing import Any, Coroutine, List, Set

from prompt_toolkit import PromptSession
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.styles import Style

from mentat.config import Config
from mentat.session import Session
from mentat.session_stream import StreamMessageSource
from mentat.terminal.loading import LoadingHandler
from mentat.terminal.output import print_stream_message
from mentat.terminal.prompt_completer import MentatCompleter
from mentat.terminal.prompt_session import MentatPromptSession
from mentat.terminal.themes import themes


class TerminalClient:
    def __init__(
        self,
        cwd: Path = Path.cwd(),
        paths: List[str] = [],
        exclude_paths: List[str] = [],
        ignore_paths: List[str] = [],
        diff: str | None = None,
        pr_diff: str | None = None,
        config: Config = Config(),
    ):
        self.cwd = cwd
        self.paths = [Path(path) for path in paths]
        self.exclude_paths = [Path(path) for path in exclude_paths]
        self.ignore_paths = [Path(path) for path in ignore_paths]
        self.diff = diff
        self.pr_diff = pr_diff
        self.config = config

        self._tasks: Set[asyncio.Task[None]] = set()
        self._should_exit = Event()
        self._stopped = Event()

    def _create_task(self, coro: Coroutine[None, None, Any]):
        """Utility method for running a Task in the background"""

        def task_cleanup(task: asyncio.Task[None]):
            self._tasks.remove(task)

        task = asyncio.create_task(coro)
        task.add_done_callback(task_cleanup)
        self._tasks.add(task)

        return task

    async def _cprint_session_stream(self):
        async for message in self.session.stream.listen():
            print_stream_message(message, themes[self.config.theme])

    async def _default_prompt_stream(self):
        self._default_prompt = ""
        async for message in self.session.stream.listen("default_prompt"):
            self._default_prompt += message.data

    async def _handle_loading_messages(self):
        loading_handler = LoadingHandler()
        async for message in self.session.stream.listen("loading"):
            loading_handler.update(message)

    async def _handle_input_requests(self):
        while True:
            input_request_message = await self.session.stream.recv("input_request")
            # TODO: Make extra kwargs like plain constants
            if input_request_message.extra.get("plain"):
                prompt_session = self._plain_session
            else:
                prompt_session = self._prompt_session

            default_prompt = self._default_prompt.strip()
            self._default_prompt = ""

            user_input = await prompt_session.prompt_async(
                handle_sigint=False, default=default_prompt
            )
            if user_input == "q":
                self._should_exit.set()
                return

            self.session.stream.send(
                user_input,
                source=StreamMessageSource.CLIENT,
                channel=f"input_request:{input_request_message.id}",
            )

    async def _listen_for_client_exit(self):
        """When the Session shuts down, it will send the client_exit signal for the client to shutdown."""
        await self.session.stream.recv(channel="client_exit")
        asyncio.create_task(self._shutdown())

    async def _listen_for_should_exit(self):
        """This listens for a user event signaling shutdown (like SigInt), and tells the session to shutdown."""
        await self._should_exit.wait()
        self.session.stream.send(
            None, source=StreamMessageSource.CLIENT, channel="session_exit"
        )

    async def _send_session_stream_interrupt(self):
        logging.debug("Sending interrupt to session stream")
        self.session.stream.send(
            "", source=StreamMessageSource.CLIENT, channel="interrupt"
        )

    # Be careful editing this function; since we use signal.signal instead of asyncio's
    # add signal handler (which isn't available on Windows), this function can interrupt
    # asyncio coroutines, potentially causing race conditions.
    def _handle_sig_int(self, sig: int, frame: FrameType | None):
        if (
            # If session is still starting up we want to quit without an error
            not self.session
            or self.session.stream.interrupt_lock.locked() is False
        ):
            if self._should_exit.is_set():
                logging.debug("Force exiting client...")
                exit(0)
            else:
                logging.debug("Should exit client...")
                self._should_exit.set()
        else:
            # We create a task here in order to avoid race conditions
            self._create_task(self._send_session_stream_interrupt())

    def _init_signal_handlers(self):
        signal.signal(signal.SIGINT, self._handle_sig_int)

    async def _run(self):
        self._init_signal_handlers()
        self.session = Session(
            self.cwd,
            self.paths,
            self.exclude_paths,
            self.ignore_paths,
            self.diff,
            self.pr_diff,
            self.config,
        )
        self.session.start()

        mentat_completer = MentatCompleter(self.session.stream)
        self._prompt_session = MentatPromptSession(
            completer=mentat_completer,
            style=Style(self.config.input_style),
            enable_suspend=True,
        )

        plain_bindings = KeyBindings()

        @plain_bindings.add("c-c")
        @plain_bindings.add("c-d")
        def _(event: KeyPressEvent):
            if event.current_buffer.text != "":
                event.current_buffer.reset()
            else:
                event.app.exit(result="q")

        self._plain_session = PromptSession[str](
            message=[("class:prompt", ">>> ")],
            style=Style(self.config.input_style),
            completer=None,
            key_bindings=plain_bindings,
            enable_suspend=True,
        )

        self._create_task(self._cprint_session_stream())
        self._create_task(self._handle_input_requests())
        self._create_task(self._handle_loading_messages())
        self._create_task(self._default_prompt_stream())
        self._create_task(self._listen_for_client_exit())
        self._create_task(self._listen_for_should_exit())

        logging.debug("Completed startup")
        await self._stopped.wait()

    async def _shutdown(self):
        logging.debug("Running shutdown")

        # Stop all background tasks
        for task in self._tasks:
            task.cancel()
        self._stopped.set()

    def run(self):
        asyncio.run(self._run())


def get_parser():
    parser = argparse.ArgumentParser(
        description="Run conversation with command line args"
    )
    parser.add_argument(
        "paths",
        nargs="*",
        default=[],
        help="List of file paths, directory paths, or glob patterns",
    )
    parser.add_argument(
        "--exclude",
        "-e",
        nargs="*",
        default=[],
        help="List of file paths, directory paths, or glob patterns to exclude",
    )
    parser.add_argument(
        "--ignore",
        "-g",
        nargs="*",
        default=[],
        help=(
            "List of file paths, directory paths, or glob patterns to ignore in"
            " auto-context"
        ),
    )
    parser.add_argument(
        "--diff",
        "-d",
        nargs="?",
        type=str,
        default=None,
        const="HEAD",
        help="A git tree-ish (e.g. commit, branch, tag) to diff against",
    )
    parser.add_argument(
        "--pr-diff",
        "-p",
        type=str,
        default=None,
        help="A git tree-ish to diff against the latest common ancestor of",
    )
    parser.add_argument(
        "--cwd", default=Path.cwd(), help="The current working directory"
    )
    Config.add_fields_to_argparse(parser)
    return parser


def run_cli():
    parser = get_parser()

    args = parser.parse_args()

    cwd = Path(args.cwd).expanduser().resolve()
    paths = args.paths
    exclude_paths = args.exclude
    ignore_paths = args.ignore
    diff = args.diff
    pr_diff = args.pr_diff

    config = Config.create(cwd, args)

    terminal_client = TerminalClient(
        cwd,
        paths,
        exclude_paths,
        ignore_paths,
        diff,
        pr_diff,
        config,
    )
    terminal_client.run()



import os
from pathlib import Path
from textwrap import dedent

import pytest

from mentat.session import Session
from tests.conftest import run_git_command


@pytest.mark.asyncio
async def test_system(mock_call_llm_api, mock_collect_user_input):
    # Create a temporary file
    temp_file_name = Path("temp.py")
    with open(temp_file_name, "w") as f:
        f.write("# This is a temporary file.")

    mock_collect_user_input.set_stream_messages(
        [
            "Add changes to the file",
            "i",
            "y",
            "q",
        ]
    )

    mock_call_llm_api.set_streamed_values([dedent("""\
        I will add a print statement.

        Steps:
        1. Add a print statement after the last line

        @@start
        {{
            "file": "{file_name}",
            "action": "replace",
            "start-line": 1,
            "end-line": 1
        }}
        @@code
        print("Hello, world!")
        @@end""".format(file_name=temp_file_name))])

    session = Session(cwd=Path.cwd(), paths=[temp_file_name])
    session.start()
    await session.stream.recv(channel="client_exit")

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = 'print("Hello, world!")'
    assert content == expected_content


@pytest.mark.asyncio
async def test_system_exits_on_exception(mock_collect_user_input):
    # if we don't catch this and shutdown properly, pytest will fail test
    # with "Task was destroyed but it is pending!"
    mock_collect_user_input.side_effect = [Exception("Something went wrong")]

    session = Session(cwd=Path.cwd())
    session.start()
    await session.stream.recv(channel="client_exit")


@pytest.mark.asyncio
async def test_interactive_change_selection(mock_call_llm_api, mock_collect_user_input):
    # Create a temporary file
    temp_file_name = Path("temp_interactive.py")
    with open(temp_file_name, "w") as f:
        f.write("# This is a temporary file for interactive test.")

    mock_collect_user_input.set_stream_messages(
        [
            "Add changes to the file",
            "i",
            "y",
            "n",
            "y",
            "q",
        ]
    )

    mock_call_llm_api.set_streamed_values([dedent("""\
        I will make three changes to the file.

        Steps:
        1. Replace the comment with print("Change 1")
        2. Add print("Change 2") after the first line
        3. Add print("Change 3") after the second line

        @@start
        {{
            "file": "{file_name}",
            "action": "replace",
            "start-line": 1,
            "end-line": 1
        }}
        @@code
        print("Change 1")
        @@end
        @@start
        {{
            "file": "{file_name}",
            "action": "insert",
            "insert-after-line": 1,
            "insert-before-line": 2
        }}
        @@code
        print("Change 2")
        @@end
        @@start
        {{
            "file": "{file_name}",
            "action": "insert",
            "insert-after-line": 2,
            "insert-before-line": 3
        }}
        @@code
        print("Change 3")
        @@end""".format(file_name=temp_file_name))])

    session = Session(cwd=Path.cwd(), paths=[temp_file_name])
    session.start()
    await session.stream.recv(channel="client_exit")

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = 'print("Change 1")\n\nprint("Change 3")'

    assert content == expected_content


# Makes sure we're properly turning the model output into correct path no matter the os
@pytest.mark.asyncio
async def test_without_os_join(mock_call_llm_api, mock_collect_user_input):
    temp_dir = "dir"
    temp_file_name = "temp.py"
    temp_file_path = Path(os.path.join(temp_dir, temp_file_name))
    os.makedirs(temp_dir, exist_ok=True)
    with open(temp_file_path, "w") as f:
        f.write("# This is a temporary file.")

    mock_collect_user_input.set_stream_messages(
        ['Replace comment with print("Hello, world!")', "y", "q"]
    )

    # Use / here since that should always be what the model outputs
    fake_file_path = temp_dir + "/" + temp_file_name
    mock_call_llm_api.set_streamed_values([dedent("""\
        I will add a print statement.
        Steps:
        1. Add a print statement after the last line
        @@start
        {{
            "file": "{file_name}",
            "action": "replace",
            "start-line": 1,
            "end-line": 1
        }}
        @@code
        print("Hello, world!")
        @@end""".format(file_name=fake_file_path))])
    session = Session(cwd=Path.cwd(), paths=[temp_file_path])
    session.start()
    await session.stream.recv(channel="client_exit")
    mock_collect_user_input.reset_mock()
    with open(temp_file_path, "r") as f:
        content = f.read()
        expected_content = 'print("Hello, world!")'
    assert content == expected_content


@pytest.mark.asyncio
async def test_sub_directory(
    temp_testbed, mock_call_llm_api, mock_collect_user_input, monkeypatch
):
    with monkeypatch.context() as m:
        m.chdir("scripts")
        file_name = "calculator.py"
        mock_collect_user_input.set_stream_messages(
            [
                "Add changes to the file",
                "y",
                "q",
            ]
        )

        mock_call_llm_api.set_streamed_values([dedent(f"""\
            Conversation

            @@start
            {{
                "file": "scripts/{file_name}",
                "action": "replace",
                "start-line": 1,
                "end-line": 50
            }}
            @@code
            print("Hello, world!")
            @@end""")])

        session = Session(cwd=temp_testbed, paths=[Path("scripts", file_name)])
        session.start()
        await session.stream.recv(channel="client_exit")

        # Check if the temporary file is modified as expected
        with open(file_name, "r") as f:
            content = f.read()
            expected_content = 'print("Hello, world!")'
        assert content == expected_content


@pytest.mark.asyncio
@pytest.mark.no_git_testbed
@pytest.mark.clear_testbed
async def test_recursive_git_repositories(temp_testbed, mock_collect_user_input):
    # Tests if a git repo inside of a git repo inside of a git repo works, from outside of a git repo

    dirs = ["outer", "git1", "git2", "inner1", "git3", "inner2"]
    total_path = Path(".")

    files = []
    for dir_name in dirs:
        file_path = total_path / (dir_name + ".txt")
        with open(file_path, "w") as f:
            f.write("")
        files.append(temp_testbed / file_path)

        total_path = total_path / dir_name
        os.mkdir(total_path)
        if dir_name.startswith("git"):
            run_git_command(total_path, "init")
    file_path = total_path / "innermost.txt"
    with open(file_path, "w") as f:
        f.write("")
    files.append(temp_testbed / file_path)

    mock_collect_user_input.set_stream_messages(
        [
            "q",
        ]
    )

    session = Session(cwd=temp_testbed, paths=[Path(".")])
    session.start()
    await session.stream.recv(channel="client_exit")

    assert set(session.ctx.code_context.include_files.keys()) == set(files)



from __future__ import annotations

import base64
import io
import os
import sys
from pathlib import Path
from timeit import default_timer
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    cast,
    overload,
)

import attr
import sentry_sdk
import tiktoken
from dotenv import load_dotenv
from openai import (
    APIConnectionError,
    AsyncAzureOpenAI,
    AsyncOpenAI,
    AsyncStream,
    AuthenticationError,
)
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionContentPartParam,
    ChatCompletionMessageParam,
)
from openai.types.chat.completion_create_params import ResponseFormat
from PIL import Image

from mentat.errors import MentatError, ReturnToUser, UserError
from mentat.session_context import SESSION_CONTEXT
from mentat.utils import mentat_dir_path

TOKEN_COUNT_WARNING = 32000


def is_test_environment():
    """Returns True if in pytest and not benchmarks"""
    benchmarks_running = os.getenv("MENTAT_BENCHMARKS_RUNNING")
    return (
        "PYTEST_CURRENT_TEST" in os.environ
        and "--benchmark" not in sys.argv
        and (not bool(benchmarks_running) or benchmarks_running == "false")
    )


def api_guard(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator that should be used on any function that calls the OpenAI API

    It does two things:
    1. Raises if the function is called in tests (that aren't benchmarks)
    2. Converts APIConnectionErrors to MentatErrors
    """

    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        assert (
            not is_test_environment()
        ), "OpenAI call attempted in non-benchmark test environment!"
        try:
            return await func(*args, **kwargs)
        except APIConnectionError:
            raise MentatError(
                "API connection error: please check your internet connection and try"
                " again."
            )

    return wrapper


# Ensures that each chunk will have at most one newline character
def chunk_to_lines(chunk: ChatCompletionChunk) -> list[str]:
    content = None
    if len(chunk.choices) > 0:
        content = chunk.choices[0].delta.content
    return ("" if content is None else content).splitlines(keepends=True)


def count_tokens(message: str, model: str, full_message: bool) -> int:
    """
    Calculates the tokens in this message. Will NOT be accurate for a full prompt!
    Use prompt_tokens to get the exact amount of tokens for a prompt.
    If full_message is true, will include the extra 4 tokens used in a chat completion by this message
    if this message is part of a prompt. You do NOT want full_message to be true for a response.
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        encoding = tiktoken.get_encoding("cl100k_base")
    return len(encoding.encode(message, disallowed_special=())) + (
        4 if full_message else 0
    )


def prompt_tokens(messages: list[ChatCompletionMessageParam], model: str):
    """
    Returns the number of tokens used by a prompt if it was sent to OpenAI for a chat completion.
    Adapted from https://platform.openai.com/docs/guides/text-generation/managing-tokens
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        encoding = tiktoken.get_encoding("cl100k_base")

    num_tokens = 0
    for message in messages:
        # every message follows <|start|>{role/name}\n{content}<|end|>\n
        # this has 5 tokens (start token, role, \n, end token, \n), but we count the role token later
        num_tokens += 4
        for key, value in message.items():
            if isinstance(value, list) and key == "content":
                value = cast(List[ChatCompletionContentPartParam], value)
                for entry in value:
                    if entry["type"] == "text":
                        num_tokens += len(encoding.encode(entry["text"]))
                    if entry["type"] == "image_url":
                        image_base64: str = entry["image_url"]["url"].split(",")[1]
                        image_bytes: bytes = base64.b64decode(image_base64)
                        image = Image.open(io.BytesIO(image_bytes))
                        size = image.size
                        # As described here: https://platform.openai.com/docs/guides/vision/calculating-costs
                        scale = min(1, 2048 / max(size))
                        size = (int(size[0] * scale), int(size[1] * scale))
                        scale = min(1, 768 / min(size))
                        size = (int(size[0] * scale), int(size[1] * scale))
                        num_tokens += 85 + 170 * ((size[0] + 511) // 512) * (
                            (size[1] + 511) // 512
                        )
            elif isinstance(value, str):
                num_tokens += len(encoding.encode(value))
            if key == "name":  # if there's a name, the role is omitted
                num_tokens -= 1  # role is always required and always 1 token
    num_tokens += 2  # every reply is primed with <|start|>assistant
    return num_tokens


@attr.define
class Model:
    name: str = attr.field()
    context_size: int = attr.field()
    input_cost: float = attr.field()
    output_cost: float = attr.field()
    embedding_model: bool = attr.field(default=False)


known_models: Dict[str, Model] = {
    "gpt-4-1106-preview": Model("gpt-4-1106-preview", 128000, 0.01, 0.03),
    "gpt-4-vision-preview": Model("gpt-4-vision-preview", 128000, 0.01, 0.03),
    "gpt-4": Model("gpt-4", 8192, 0.03, 0.06),
    "gpt-4-32k": Model("gpt-4-32k", 32768, 0.06, 0.12),
    "gpt-4-0613": Model("gpt-4-0613", 8192, 0.03, 0.06),
    "gpt-4-32k-0613": Model("gpt-4-32k-0613", 32768, 0.06, 0.12),
    "gpt-4-0314": Model("gpt-4-0314", 8192, 0.03, 0.06),
    "gpt-4-32k-0314": Model("gpt-4-32k-0314", 32768, 0.06, 0.12),
    "gpt-3.5-turbo-1106": Model("gpt-3.5-turbo-1106", 16385, 0.001, 0.002),
    "gpt-3.5-turbo": Model("gpt-3.5-turbo", 16385, 0.001, 0.002),
    "gpt-3.5-turbo-0613": Model("gpt-3.5-turbo-0613", 4096, 0.0015, 0.002),
    "gpt-3.5-turbo-16k-0613": Model("gpt-3.5-turbo-16k-0613", 16385, 0.003, 0.004),
    "gpt-3.5-turbo-0301": Model("gpt-3.5-turbo-0301", 4096, 0.0015, 0.002),
    "text-embedding-ada-002": Model(
        "text-embedding-ada-002", 8191, 0.0001, 0, embedding_model=True
    ),
}


def model_context_size(model: str) -> Optional[int]:
    if model not in known_models:
        return None
    else:
        return known_models[model].context_size


def model_price_per_1000_tokens(model: str) -> Optional[tuple[float, float]]:
    """Returns (input, output) cost per 1000 tokens in USD"""
    if model not in known_models:
        return None
    else:
        return known_models[model].input_cost, known_models[model].output_cost


def get_max_tokens() -> int:
    session_context = SESSION_CONTEXT.get()
    stream = session_context.stream
    config = session_context.config

    context_size = model_context_size(config.model)
    maximum_context = config.maximum_context

    if context_size is not None and maximum_context is not None:
        return min(context_size, maximum_context)
    elif context_size is not None:
        return context_size
    elif maximum_context is not None:
        return maximum_context
    else:
        maximum_context = 4096
        # This attr has a converter from str to int
        config.maximum_context = str(maximum_context)
        stream.send(
            f"Context size for {config.model} is not known. Set maximum-context"
            " with `/config maximum_context <value>`. Using a default value of"
            f" {maximum_context}.",
            style="error",
        )
        return maximum_context


def is_context_sufficient(tokens: int) -> bool:
    ctx = SESSION_CONTEXT.get()

    max_tokens = get_max_tokens()
    if max_tokens - tokens < ctx.config.token_buffer:
        ctx.stream.send(
            f"The context size is limited to {max_tokens} tokens and your current"
            f" request uses {tokens} tokens. Please use `/exclude` to remove"
            " some files from context or `/clear` to reset the conversation",
            style="error",
        )
        return False

    return True


class LlmApiHandler:
    """Used for any functions that require calling the external LLM API"""

    def initialize_client(self):
        ctx = SESSION_CONTEXT.get()

        if not load_dotenv(mentat_dir_path / ".env"):
            load_dotenv()
        key = os.getenv("OPENAI_API_KEY")
        base_url = os.getenv("OPENAI_API_BASE")
        azure_key = os.getenv("AZURE_OPENAI_KEY")
        azure_endpoint = os.getenv("AZURE_OPENAI_ENDPOINT")

        # We don't have any use for a synchronous client, but if we ever do we can easily make it here
        if azure_endpoint and azure_key:
            ctx.stream.send("Using Azure OpenAI client.", color="cyan")
            self.async_client = AsyncAzureOpenAI(
                api_key=azure_key,
                api_version="2023-12-01-preview",
                azure_endpoint=azure_endpoint,
            )
        else:
            if not key:
                if not base_url:
                    raise UserError(
                        "No OpenAI api key detected.\nEither place your key into a .env"
                        " file or export it as an environment variable."
                    )
                # If they set the base_url but not the key, they probably don't need a key, but the client requires one
                key = "fake_key"
            self.async_client = AsyncOpenAI(api_key=key, base_url=base_url)

        try:
            self.async_client.models.list()  # Test the key
        except AuthenticationError as e:
            raise UserError(f"API gave an Authentication Error:\n{e}")

    @overload
    async def call_llm_api(
        self,
        messages: list[ChatCompletionMessageParam],
        model: str,
        stream: Literal[True],
        response_format: ResponseFormat = ResponseFormat(type="text"),
    ) -> AsyncIterator[ChatCompletionChunk]: ...

    @overload
    async def call_llm_api(
        self,
        messages: list[ChatCompletionMessageParam],
        model: str,
        stream: Literal[False],
        response_format: ResponseFormat = ResponseFormat(type="text"),
    ) -> ChatCompletion: ...

    @api_guard
    async def call_llm_api(
        self,
        messages: list[ChatCompletionMessageParam],
        model: str,
        stream: bool,
        response_format: ResponseFormat = ResponseFormat(type="text"),
    ) -> ChatCompletion | AsyncIterator[ChatCompletionChunk]:
        session_context = SESSION_CONTEXT.get()
        config = session_context.config
        cost_tracker = session_context.cost_tracker

        # Confirm that model has enough tokens remaining.
        tokens = prompt_tokens(messages, model)
        if not is_context_sufficient(tokens):
            raise ReturnToUser()

        start_time = default_timer()
        with sentry_sdk.start_span(description="LLM Call") as span:
            span.set_tag("model", model)
            # OpenAI's API is bugged; when gpt-4-vision-preview is used, including the response format
            # at all returns a 400 error. Additionally, gpt-4-vision-preview has a max response of 30 tokens by default.
            # Until this is fixed, we have to use this workaround.
            if model == "gpt-4-vision-preview":
                response = await self.async_client.chat.completions.create(
                    model=model,
                    messages=messages,
                    temperature=config.temperature,
                    stream=stream,
                    max_tokens=4096,
                )
            else:
                # This makes it slightly easier when using the litellm proxy or models outside of OpenAI
                if response_format["type"] == "text":
                    response = await self.async_client.chat.completions.create(
                        model=model,
                        messages=messages,
                        temperature=config.temperature,
                        stream=stream,
                    )
                else:
                    response = await self.async_client.chat.completions.create(
                        model=model,
                        messages=messages,
                        temperature=config.temperature,
                        stream=stream,
                        response_format=response_format,
                    )

        # We have to cast response since pyright isn't smart enough to connect
        # the dots between stream and the overloaded create function
        if not stream:
            time_elapsed = default_timer() - start_time
            response_tokens = count_tokens(
                cast(ChatCompletion, response).choices[0].message.content or "",
                model,
                full_message=False,
            )
            cost_tracker.log_api_call_stats(
                tokens, response_tokens, model, time_elapsed
            )
        else:
            cost_tracker.last_api_call = ""
            response = cost_tracker.response_logger_wrapper(
                tokens, cast(AsyncStream[ChatCompletionChunk], response), model
            )

        return response

    @api_guard
    async def call_embedding_api(
        self, input_texts: list[str], model: str = "text-embedding-ada-002"
    ) -> list[list[float]]:
        response = await self.async_client.embeddings.create(
            input=input_texts, model=model
        )
        return [embedding.embedding for embedding in response.data]

    @api_guard
    async def call_whisper_api(self, audio_path: Path) -> str:
        audio_file = open(audio_path, "rb")
        transcript = await self.async_client.audio.transcriptions.create(
            model="whisper-1",
            file=audio_file,
        )
        return transcript.text



import os
from timeit import default_timer as timer

import fire
import openai
import tiktoken
from dotenv import load_dotenv


def get_tokens_from_message(message: str) -> int:
    return len(tiktoken.encoding_for_model("gpt-4").encode(message))


# Set up the OpenAI API
load_dotenv("mentat/.env")
openai.api_key = os.getenv("OPENAI_API_KEY")


def run(prompt, model: str = "gpt-4-0314") -> None:
    messages = [
        {"role": "system", "content": "be a helpful assistant"},
        {"role": "user", "content": prompt},
    ]

    start = timer()
    response = openai.ChatCompletion.create(
        model=model,
        messages=messages,
        temperature=0.5,
        stream=True,
    )

    started = False
    message = []
    for chunk in response:
        if not started:
            time_to_first_token = timer() - start
            started = True
        delta = chunk.choices[0].delta
        content = delta.get("content", None)
        if content is not None:
            message.append(content)
            print(content, end="", flush=True)
    time_to_last_token = timer() - start
    message = "".join(message)
    token_count = get_tokens_from_message(message)

    print("\n")
    print(f"Token count: {token_count}")
    print(f"Time to first token: {time_to_first_token}")
    print(f"Tokens/second: {token_count / time_to_last_token}")
    print(f"Time to last token: {time_to_last_token}")


if __name__ == "__main__":
    fire.Fire(run)



import argparse
from pathlib import Path
from textwrap import dedent

import pytest

import mentat.config
from mentat.config import Config, config_file_name
from mentat.parsers.replacement_parser import ReplacementParser


@pytest.fixture
def mock_config_errors(mocker):
    errors = []
    mocker.patch.object(Config, "error", lambda self, message: errors.append(message))
    return errors


@pytest.mark.asyncio
async def test_config_creation():
    "This test verifies the Config adds the parameters to the argparse object."
    "Those take precedence over the config files and the project config takes"
    "precedence over the user config."
    parser = argparse.ArgumentParser()
    Config.add_fields_to_argparse(parser)
    args = parser.parse_args(
        [
            "--model",
            "model",
            "--temperature",
            "0.2",
            "--maximum-context",
            "1",
            "-a",
            "2000",
        ]
    )
    assert args.model == "model"
    assert args.temperature == 0.2
    assert args.maximum_context == "1"
    assert args.parser is None
    assert args.auto_context_tokens == 2000

    with open(config_file_name, "w") as project_config_file:
        project_config_file.write(dedent("""\
        {
            "input_style": [[ "project", "yes" ]]
        }"""))

    mentat.config.user_config_path = Path(str(config_file_name) + "1")
    with open(mentat.config.user_config_path, "w") as user_config_file:
        user_config_file.write(dedent("""\
        {
            "model": "test",
            "parser": "replacement",
            "input_style": [[ "user", "yes" ]]
        }"""))

    config = Config.create(Path.cwd(), args)

    assert config.model == "model"
    assert config.temperature == 0.2
    assert config.maximum_context == 1
    assert type(config.parser) == ReplacementParser
    assert config.auto_context_tokens == 2000
    assert config.input_style == [["project", "yes"]]


@pytest.mark.asyncio
async def test_invalid_config(mock_config_errors):
    # If invalid config file is found, it should use next config
    with open(config_file_name, "w") as project_config_file:
        project_config_file.write(dedent("""\
        {
            "model": "project",
            "format": "I have a trailing comma",
        }"""))

    mentat.config.user_config_path = Path(str(config_file_name) + "1")
    with open(mentat.config.user_config_path, "w") as user_config_file:
        user_config_file.write(dedent("""\
        {
            "model": "test",
            "foobar": "Not a real setting"
        }"""))

    config = Config.create(cwd=Path.cwd())
    assert (
        mock_config_errors[0]
        == "Warning: Config .mentat_config.json1 contains unrecognized setting: foobar"
    )
    assert (
        "contains invalid json; ignoring user configuration file"
        in mock_config_errors[1]
    )
    assert config.model == "test"
